"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""


class AvgRecord:
    """
    用于保存累积均值的类
    avg = AvgItem()
    avg += 1 # avg.update(1)
    avg += 2
    avg += 3

    avg.item = 3 #(last item)
    avg.avg = 2 #(average item)
    avg.sum = 6
    """
    def __init__(self, weight=1) -> None:
        super().__init__()
        self.sum = 0
        self.weight = weight
        self.count = 0
        self.item = None

    def __add__(self, other):
        self.update(other)

    def update(self, other, weight=1):
        self.sum += other * self.weight
        self.count += self.weight
        self.item = other

    @property
    def avg(self):
        return self.sum / self.count

    def __repr__(self) -> str:
        return str("{}({})".format(self.item, self.avg))



"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""

from torch.optim.adadelta import Adadelta
from torch.optim.adagrad import Adagrad
from torch.optim.adam import Adam
from torch.optim.adamax import Adamax
from torch.optim.adamw import AdamW
from torch.optim.asgd import ASGD
from torch.optim.lbfgs import LBFGS
from torch.optim.optimizer import Optimizer
from torch.optim.rmsprop import RMSprop
from torch.optim.rprop import Rprop
from torch.optim.sgd import SGD
from torch.optim.sparse_adam import SparseAdam
from torch.optim import adam,adamw,adamax,adagrad,adadelta,asgd,sparse_adam,sgd,lr_scheduler,lbfgs,optimizer,rmsprop,rprop



"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""

import random

import numpy as np
import torch


def fix_seed(self, seed=10):
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.random.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True

    return {
        "numpy": np.random.get_state(),
        "torch": torch.random.get_rng_state(),
        "torch.cuda": torch.cuda.get_rng_state() if torch.cuda.is_available() else None,
        "random": random.getstate(),
    }

def get_state():
    return {
        "numpy": np.random.get_state(),
        "torch": torch.random.get_rng_state(),
        "torch.cuda": torch.cuda.get_rng_state() if torch.cuda.is_available() else None,
        "random": random.getstate(),
    }

def set_state(state_dict):
    np.random.set_state(state_dict["numpy"])
    torch.random.set_rng_state(state_dict["torch"])
    if torch.cuda.is_available():
        if state_dict["torch.cuda"]:
            torch.cuda.set_rng_state(state_dict["torch.cuda"])
        else:
            import warnings
            warnings.warn("Don't have torch.cuda random state")
    random.setstate(state_dict["random"])





"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""
import json
import os
import re

import torch

from ..utils.generel_util import listdir_by_time


class Saver:
    _ckpt_fn_templete = "{}{:06d}.ckpt"
    _model_fn_templete = "model.{}{:06d}.pth"
    re_fn = re.compile("^[0-9]{7}\.ckpt$")
    re_keep_fn = re.compile("^keep.[0-9]{7}\.ckpt$")

    def __init__(self, ckpt_dir, max_to_keep=3, ):
        self.info = {}
        self.ckpt_dir = ckpt_dir
        self.max_to_keep = max_to_keep
        os.makedirs(ckpt_dir, exist_ok=True)

    def _build_checkpoint_name(self, epoch, replacement: bool, lasting):
        i = 0

        def build_fn():
            fn = Saver._ckpt_fn_templete.format(i, epoch)
            if lasting:
                fn = "keep.{}".format(fn)
            return fn

        absfn = os.path.join(self.ckpt_dir, build_fn())
        while replacement == False and os.path.exists(absfn):
            i += 1
            if i >= 9:
                if lasting:
                    kfns = self.find_keypoints()
                else:
                    kfns = self.find_checkpoints()
                return kfns[-1]

            absfn = os.path.join(self.ckpt_dir, build_fn())
        return absfn

    def _build_model_name(self, epoch):
        fn = os.path.join(self.ckpt_dir, Saver._model_fn_templete.format(0, epoch))
        return fn

    def find_models(self):
        fs = listdir_by_time(self.ckpt_dir)  # 按创建时间排序
        fs = [i for i in fs if i.endswith(".pth")]
        return fs

    def find_keypoint(self, epoch, all=False):
        """
        find all "keeped" checkpoint saved in the save dir.
        :return:
        """
        fs = listdir_by_time(self.ckpt_dir)  # 按创建时间排序
        re_fn = re.compile("^keep.[0-9]{:06d}\.ckpt$".format(epoch))
        fs = [os.path.join(self.ckpt_dir, i) for i in fs if re.search(re_fn, i) is not None]
        if all:
            return fs
        if len(fs) > 0:
            return fs[0]  # 返回最后保存的那一个
        return None

    def find_keypoint_info(self, epoch, all=False):
        fns = self.find_keypoint(epoch, all)
        if all:
            return ["{}.json".format(fn) for fn in fns]
        else:
            if fns is None:
                return None
            return "{}.json".format(fns)

    def find_keypoints(self):
        """
        find all "keeped" checkpoint saved in the save dir.
        :return:
        """
        fs = listdir_by_time(self.ckpt_dir)  # 按创建时间排序
        fs = [os.path.join(self.ckpt_dir, i) for i in fs if re.search(Saver.re_keep_fn, i) is not None]
        return fs

    def find_checkpoint(self, epoch, all=False):
        """
        find checkpoints saved in [epoch]
        :param epoch:  int
        :param all: 是否返回多个如果发现了多个保存在同一个epoch的checkpoint,
            default False
        :return:  if 'all' is True, return all checkpoints in that epoch
            else, return the last created checkpoint
        """
        fs = listdir_by_time(self.ckpt_dir)  # 按创建时间排序
        re_fn = re.compile("^[0-9]{:06d}\.ckpt$".format(epoch))
        fs = [os.path.join(self.ckpt_dir, i) for i in fs if re.search(re_fn, i) is not None]
        if all:
            return fs
        if len(fs) > 0:
            return fs[0]  # 返回最后保存的那一个
        else:
            return None

    def find_checkpoints(self):
        """
        find all checkpoint saved in the save dir.
        :return:
        """
        fs = listdir_by_time(self.ckpt_dir)  # 按创建时间排序
        fs = [os.path.join(self.ckpt_dir, i) for i in fs if re.search(Saver.re_fn, i) is not None]
        return fs

    def find_checkpoint_info(self, epoch, all=False):
        fns = self.find_checkpoint(epoch, all)
        if all:
            return ["{}.json".format(fn) for fn in fns]
        else:
            if fns is None:
                return None
            return "{}.json".format(fns)

    def _check_max_checkpoint(self):
        fs = self.find_checkpoints()
        while len(fs) > self.max_to_keep:
            self.check_remove(fs[-1])
            fs.pop()

    def save_model(self, val, state_dict, extra_info: dict = None):
        if self._check_isint_or_str(val):
            fn = self._build_model_name(val)
        else:
            fn = val
        json_fn = "{}.json".format(fn)
        torch.save(state_dict, fn)
        with open(json_fn, "w", encoding="utf-8") as w:
            json.dump(extra_info, w, indent=2)

        return fn

    def save_checkpoint(self, val, state_dict, extra_info: dict = None, replacement: bool = False, lasting=False):
        if self._check_isint_or_str(val):
            fn = self._build_checkpoint_name(val, replacement, lasting)
        else:
            fn = val

        json_fn = "{}.json".format(fn)
        torch.save(state_dict, fn)
        with open(json_fn, "w", encoding="utf-8") as w:
            json.dump(extra_info, w, indent=2)
            self._check_max_checkpoint()

        return fn

    def save_keypoint(self, val, state_dict, extra_info: dict = None, replacement: bool = False):
        return self.save_checkpoint(val, state_dict, extra_info, replacement, True)

    def _check_isint_or_str(self, val):
        assert type(val) in {int, str}
        if isinstance(val, int):
            return True
        return False

    def load_model(self, val):
        if self._check_isint_or_str(val):
            fn = self._build_model_name(val)
        else:
            fn = val
        return torch.load(fn)

    def load_model_info(self, val):
        if self._check_isint_or_str(val):
            fn = self._build_model_name(val)
        else:
            fn = val
        json_fn = "{}.json".format(fn)
        with open(json_fn, "r", encoding="utf-8") as r:
            return json.load(r)

    def load_keypoint(self, val):
        """

        :param val: epoch number[int] or file path[str]
        :return:  None or torch.load() result
        """
        if self._check_isint_or_str(val):
            fn = self.find_keypoint(val)
        else:
            fn = val
        if fn is not None and os.path.exists(fn):
            return torch.load(fn)
        return None

    def load_keypoint_info(self, val):
        """

        :param val: epoch number[int] or file path[str]
        :return:  None or torch.load() result
        """
        if self._check_isint_or_str(val):
            fn = self.find_keypoint(val)
        else:
            fn = val
        if fn is not None:
            json_fn = "{}.json".format(fn)
            with open(json_fn, "r", encoding="utf-8") as r:
                return json.load(r)
        return None

    def load_checkpoint(self, val, byindex=False):
        """

        :param val: epoch number[int] or file path[str]
        :return:  None or torch.load() result
        """
        assert not (not isinstance(val, int) and byindex), "If byindex = True, val must be an index(int)."
        if byindex:
            fn = self.find_checkpoints()[val]
        elif self._check_isint_or_str(val):
            fn = self.find_checkpoint(val)
        else:
            fn = val
        if fn is not None and os.path.exists(fn):
            return torch.load(fn)
        return None

    def load_checkpoint_info(self, val, byindex=False):
        assert not (not isinstance(val, int) and byindex), "If byindex = True, val must be an index(int)."
        if byindex:
            fn = self.find_checkpoints()[val]
        elif self._check_isint_or_str(val):
            fn = self.find_checkpoint(val)
        else:
            fn = val
        if fn is not None:
            json_fn = "{}.json".format(fn)
            with open(json_fn, "r", encoding="utf-8") as r:
                return json.load(r)
        return None

    def check_remove(self, fn, with_json=True):
        if os.path.exists(fn):
            os.remove(fn)
        if with_json:
            jfn = "{}.json".format(fn)
            if os.path.exists(jfn):
                os.remove(jfn)

    def clear_models(self):
        fns = self.find_models()
        for i in fns:
            self.check_remove(i)

    def clear_checkpoints(self):
        for i in self.find_checkpoints():
            self.check_remove(i)

    def clear_keypoints(self):
        for i in self.find_keypoints():
            self.check_remove(i)

    def summary(self, detail=False):
        print("checkpoints:")
        print(" || ".join(self.find_checkpoints()))
        print("keppoints:")
        print(" || ".join(self.find_keypoints()))
        print("models:")
        print(" || ".join(self.find_models()))



"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.


"""

doc = """
Usage:
 thexp init
 thexp config -l
 thexp config --global <name> <value> 
 thexp config --global -u <name>
 thexp config --global -l
 thexp config --local <name> <value>
 thexp config --local -u <name>
 thexp config --local -l
 thexp config --global -e
 thexp config --local -e
"""
# 检测是否init，没有报错 fatal: unable to read config file '.git/config': No such file or directory
from docopt import docopt
from thexp.frame.experiment import globs
from thexp import __VERSION__
import os
def init():
    init_dir = os.path.join(os.getcwd(),".thexp")
    os.makedirs(init_dir,exist_ok=True)

    if globs.local_fn != None:
        print("ok.")

def update(mode,name,val):
    globs.update(mode,name,val)

def unset(mode,name):
    globs.unset(mode,name)

def list(globa=False,local=False):
    globs.list_config(globa,local)

arguments = docopt(doc, version=__VERSION__)

if arguments["init"]:
    init()
elif arguments["config"]:
    if arguments["--global"]:
        if arguments["-l"]:
            list(globa=True)
        elif arguments["<name>"]:
            if arguments["-u"]:
                globs.unset("global",arguments["<name>"])
            else:
                globs.update("global",arguments["<name>"],arguments["<value>"])
    elif arguments["--local"]:
        if arguments["-l"]:
            list(local=True)
        elif arguments["<name>"]:
            if arguments["-u"]:
                globs.unset("local", arguments["<name>"])
            else:
                globs.update("local", arguments["<name>"], arguments["<value>"])
    elif arguments["-l"]:
        list(True,True)
else:
    print(arguments)

exit(0)


"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""
import pprint as pp
from collections import OrderedDict
from typing import Any

import fire
import torch


class BaseParams:
    def __init__(self):
        self._param_dict = OrderedDict()
        self._exp_name = None

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            super().__setattr__(name, value)
        else:
            self._param_dict[name] = value

    def __setitem__(self, key, value):
        key = str(key)
        self.__setattr__(key, value)

    def __getattr__(self, item):
        if item not in self._param_dict:
            raise AttributeError(item)
        return self._param_dict[item]

    def __repr__(self):
        return "{}".format(self.__class__.__name__) + pp.pformat([(k, v) for k, v in self._param_dict.items()])

    def _can_in_dir_name(self, obj):
        for i in [int, float, str, bool]:
            if isinstance(obj, i):
                return True
        if isinstance(obj, torch.Tensor):
            if len(obj.shape) == 0:
                return True
        return False

    def build_exp_name(self, *names, prefix="", sep="_", ignore_mode="add"):
        prefix = prefix.strip()
        res = []
        if len(prefix) != 0:
            res.append(prefix)
        if ignore_mode == "add":
            for name in names:
                if hasattr(self, name):
                    obj = getattr(self, name)
                    if self._can_in_dir_name(obj):
                        res.append("{}={}".format(name, obj))
                else:
                    res.append(name)

        elif ignore_mode == "del":
            for name in names:
                if hasattr(self, name):
                    obj = getattr(self, name)
                    if self._can_in_dir_name(obj):
                        res.append("{}={}".format(name, obj))
        else:
            assert False

        self._exp_name = sep.join(res)
        return self._exp_name

    def get_exp_name(self):
        assert self._exp_name is not None, "run build_exp_name() before get_exp_name()"
        return self._exp_name

    def from_args(self):
        def func(**kwargs):
            for k, v in kwargs.items():
                self[k] = v

        fire.Fire(func)


class Params(BaseParams):
    def __init__(self):
        super().__init__()
        self.epoch = 10
        self.eidx = 1
        self.idx = 0
        self.global_step = 0
        self.device = "cuda:0" if torch.cuda.is_available() else "cpu"


if __name__ == '__main__':
    print(Params())



"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.

用于
"""

from ..frame import Params,Meter
import torch
from . import accuracy as acc

class classify_logits:
    def test_eval_logic(self, dataloader, param: Params):
        with torch.no_grad():
            count_dict = Meter()
            for xs, labels in dataloader:
                xs, labels = xs.to(self.device), labels.to(self.device)
                preds = self.predict(xs)
                total, topk_res = acc.classify(preds, labels, topk=param.topk)
                count_dict["total"] += total
                for i, topi_res in zip(param.topk, topk_res):
                    count_dict[i] += topi_res
        return count_dict



"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""
import os
import pickle

from ..utils import random


class RndManager:
    """用于控制训练过程中的随机种子，保证数据集/"""
    STATE_DATESET = "dataset" # 推荐在加载数据集前用，保证每次随机切分的数据集完全相同
    STATE_MODEL = "model" # 推荐在初始化模型时候用，保证模型初始化参数完全相同
    STATE_TRAIN_BEGIN = "begin" # 推荐在训练开始前用，保证每次训练的反向传播更新过程完全相同
    STATE_BREAK = "break"

    def __init__(self, save_dir):
        self.save_dir = save_dir

    def save_rnd_state(self, name):
        stt = random.get_state()
        with open(self._build_state_name(name), "wb") as f:
            pickle.dump(stt, f)

    def have_rnd_state(self, name):
        return os.path.exists(self._build_state_name(name))

    def get_rnd_state(self, name):
        if not self.have_rnd_state(name):
            return None
        with open(self._build_state_name(name), "rb") as f:
            return pickle.load(f)

    def load_rnd_state(self, name):
        stt = self.get_rnd_state(name)
        if stt is not None:
            random.set_state(stt)
            return True
        return False

    def del_rnd_state(self,name):
        if self.have_rnd_state(name):
            os.remove(self._build_state_name(name))

    def _build_state_name(self, name):
        return os.path.join(self.save_dir, "{}.rnd".format(name))



"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""
from typing import Any


class Self:
    def __getattr__(self, item):
        return item

    def __getitem__(self, item):
        return item


class NoneItem:
    """
    可以和任何元素做加减乘除，在计算的过程中根据运算
    该类相当于 零元（0）或单位元（1）
    """

    @staticmethod
    def clone(x):
        pass

    def __eq__(self, o: object) -> bool:
        return o is None

    def __add__(self, other):
        return 0 + other

    def __mul__(self, other):
        return 1 * other

    def __sub__(self, other):
        return 0 - other

    def __truediv__(self, other):
        return 1 / other

    def __floordiv__(self, other):
        return other

    def __repr__(self):
        return "NoneItem()"

    def __lt__(self, other):
        return True

    def __le__(self, other):
        return True

    def __ge__(self, other):
        return True

    def __gt__(self, other):
        return True

    def __cmp__(self, other):
        return True

    def __ne__(self, other):
        return other is not None


class AvgItem:
    """
    用于保存累积均值的类
    avg = AvgItem()
    avg += 1 # avg.update(1)
    avg += 2
    avg += 3

    avg.item = 3 #(last item)
    avg.avg = 2 #(average item)
    avg.sum = 6
    """
    def __init__(self, weight=1) -> None:
        super().__init__()
        self._sum = 0
        self._weight = weight
        self._count = 0
        self._item = 0

    def __add__(self, other):
        self.update(other)
        return self

    def update(self, other, weight=1):
        self._sum += other * self._weight
        self._count += self._weight
        self._item = other

    @property
    def avg(self):
        if self._count == 0:
            return 0
        return self._sum / self._count

    def __repr__(self) -> str:
        if self._item == self.avg:
            return str(self._item)
        return str("{}({})".format(self._item, self.avg))

    def __getattribute__(self, name: str) -> Any:
        return super().__getattribute__(name)

    def __getattr__(self, item):
        return getattr(self._item, item)

    def __format__(self, format_spec):
        return "{{:{}}}({{:{}}})".format(format_spec,format_spec).format(self._item, self.avg)

    def __getitem__(self, item):
        return self._item[item]


"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""
import atexit
import hashlib
import inspect
import json
# from .trainer import BaseTrainer
import os
import sys
from collections import Iterable, defaultdict
from datetime import datetime
from datetime import timedelta
from typing import Any

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from ..utils import generel_util as gu
from ..utils.pickledb import PickleDB

__all__ = ["Experiment", "ExperimentViewer"]
import pprint as pp

from ..base_classes.tree import tree


class __Global:
    """global var"""

    @property
    def glob_fn(self):
        return self._glob_fn

    @property
    def local_fn(self):
        if self._local_fn is None:
            self._local_fn = self.check_local_dir_exists()
            if self._local_fn is not None:
                self._local_info = self.load(self.local_fn)
            else:
                sys.stderr.write("fatal: unable to read config file '.thexp/config.json': No such file or directory\n")
                exit(1)
        return self._local_fn

    def __init__(self):
        exp_dir = gu.home_dir()
        self._glob_fn = os.path.join(exp_dir, "config.json")
        self._local_fn = self.check_local_dir_exists()

        self._glob_info = self.load(self.glob_fn)
        self._local_info = self.load(self._local_fn)

    def list_config(self, glob=True, local=True):
        if glob:
            print("global:")
            pp.pprint(self.load(self.glob_fn))

        if local:
            print("local:")
            pp.pprint(self.load(self.local_fn))

    def update(self, mode, name, val):
        if mode == "global":
            self._glob_info[name] = val
            self.dump(self.glob_fn, self._glob_info)
        elif mode == "local":
            self._local_info[name] = val
            self.dump(self.local_fn, self._local_info)

    def unset(self, mode, name):
        if mode == "global":
            if name in self._glob_info:
                self._glob_info.pop(name)
                self.dump(self.glob_fn, self._glob_info)
        elif mode == "local":
            if name in self._local_info:
                self._local_info.pop(name)
                self.dump(self.local_fn, self._local_info)

    def check_local_dir_exists(self):
        # import os
        cur = os.getcwd()
        while len(cur):
            if os.path.exists(os.path.join(cur, ".thexp")):
                local_fn = os.path.join(cur, ".thexp", "config.json")
                if local_fn != self.glob_fn:
                    return local_fn

            ncur, _ = os.path.split(cur)
            if cur != ncur:
                cur = ncur
            else:
                break
        return None

    def load(self, fn):
        if fn is None:
            return {}

        if not os.path.exists(fn):
            self.dump(fn, {})
            return {}
        with open(fn, "r", encoding="utf-8") as f:
            return json.load(f)

    def dump(self, fn, obj):
        with open(fn, "w", encoding="utf-8") as f:
            json.dump(obj, f, indent=2)

    def __getitem__(self, item):
        if item.startswith("_"):
            raise AttributeError(item)
        if item in self._local_info:
            return self._local_info[item]
        elif item in self._glob_info:
            return self._glob_info[item]
        raise AttributeError(item)

globs = __Global()


class Experiment:
    _instance = None

    @property
    def dbfile(self):
        dbfile = os.path.join(self.dbpath, self.dbfilename)
        return os.path.normpath(dbfile)

    def __init__(self, exps_dir=None, dbfilename=None, dbpath=None):
        if dbfilename is None:
            dbfilename = "thexp.json"
        self.dbfilename = dbfilename
        if dbpath is None:
            dbpath = gu.home_dir()
        self.dbpath = dbpath

        self.initial_exp_info()

        if exps_dir is None:
            main_dir = os.path.abspath(os.path.split(sys.argv[0])[0])
            exps_dir = os.path.join(main_dir, "experiment")
        self.exp_dir = exps_dir
        os.makedirs(self.exp_dir, exist_ok=True)
        self.glob = globs
        self.db = PickleDB(self.dbfile, auto_dump=True)

        self.event_handler = ExperimentEventHandler()
        self.observer = Observer()
        self.snap_member = []
        self.keycode(sys.modules["__main__"])
        atexit.register(self.end_exp)
        # self.exithook = gu.exithook()

    def initial_exp_info(self):
        exp_info = tree()
        exp_info["start_time"] = gu.curent_date()
        self.exp_info = exp_info

    def _cacu_hash(self, str):
        hl = hashlib.md5()
        hl.update(str.encode(encoding='utf-8'))
        return hl.hexdigest()[:16]

    def hold_exp_part(self, dir, exts: Iterable = (), observe=True):
        path = os.path.join(self.exp_dir, dir)
        os.makedirs(path, exist_ok=True)
        if observe:
            for i in exts:
                self.event_handler.add_prefix(i)
            self.observer.schedule(self.event_handler, path, recursive=True)
        return path

    def add_event_listener(self, func, exts: Iterable = ()):
        assert len(list(exts)) != 0
        self.event_handler.add_event_listener(func, exts=exts)

    def start_exp(self):
        self.observer.start()
        for member in self.snap_member:
            name = member.__name__
            code = inspect.getsource(member)
            fn = inspect.getfile(member)
            hash = gu.string_hash(code)

            i = 0

            refn = os.path.split(fn)[1]

            snap_path = os.path.join(self.hold_exp_part("code", observe=False), "{}-{}-{}".format(refn, name, i))

            while os.path.exists(snap_path) and gu.file_hash(snap_path) != hash:
                i += 1
                snap_path = os.path.join(self.hold_exp_part("code", observe=False), "{}-{}-{}".format(refn, name, i))

            self.snapshot(code, snap_path)
            self.exp_info["snap"][name] = dict(
                fn=fn,
                hash=hash,
                snap=snap_path,
            )

        self.exp_info["args"] = sys.argv
        self.exp_info["files"] = self.event_handler.file_dict

    def end_exp(self):
        self.exp_info["end_time"] = gu.curent_date()
        exp_dir = self.hold_exp_part("exp", observe=False)
        datekey = gu.curent_date()
        with open(os.path.join(exp_dir, "{}.json".format(datekey)), "w", encoding="utf-8") as w:
            json.dump(self.exp_info, w, indent=2)

        self.db.set(datekey, self.exp_info)
        print("\nexp info saved in {},key='{}'".format(self.dbfile, datekey))

    def snapshot(self, code, path):
        with open(path, "w", encoding="utf-8") as w:
            w.write(code)

    def keycode(self, member=None):
        if member is not None:
            self.snap_member.append(member)
            return

        def func(member):
            self.snap_member.append(member)
            return member

        return func

    def summary(self):
        pp.pprint(self.exp_info)

    def __getattr__(self, item):
        return self.glob[item]


class ExperimentEventHandler(FileSystemEventHandler):
    """用来获取试验过程中的模型保存和日志输出的变化 TODO """

    def __init__(self):
        self.file_dict = tree()
        self.ext_set = set()
        self.listener_dict = defaultdict(list)

    def add_prefix(self, ext: str):
        assert ext.startswith(".")
        self.ext_set.add(ext.lower())

    def add_event_listener(self, func, exts: Iterable):
        """
        if file created/modified/deleted and its ext in 'exts', func(file, mode) will be called.
        :param func:  functions with 2 args
        :param exts:
        :return:
        """
        for ext in exts:
            assert ext.startswith("."), "ext must have '.', like '.png', '.log'"
            ext = ext.lower()
            self.listener_dict[ext].append(func)

    def update(self, fn, mode):
        ext = os.path.splitext(fn)[1].lower()

        if ext in self.listener_dict:
            for func in self.listener_dict[ext]:
                func(fn, mode)

        if ext in self.ext_set:
            self.file_dict[fn]["mode"] = mode
            if os.path.exists(fn):
                if os.path.exists("{}.json".format(fn)):
                    self.file_dict[fn]["info"] = "{}.json".format(fn)
                self.file_dict[fn]["hash"] = gu.file_atime_hash(fn)
                self.file_dict[fn]["atime"] = gu.file_atime2date(fn)
                self.file_dict[fn]["exist"] = True
            else:
                self.file_dict[fn]["exist"] = False

    def on_created(self, event):
        super(ExperimentEventHandler, self).on_created(event)
        if not event.is_directory:
            fn = event.src_path
            self.update(fn, "created")

    def on_deleted(self, event):
        super(ExperimentEventHandler, self).on_deleted(event)
        if not event.is_directory:
            fn = event.src_path
            self.update(fn, "deleted")

    def on_modified(self, event):
        super(ExperimentEventHandler, self).on_modified(event)
        if not event.is_directory:
            fn = event.src_path
            self.update(fn, "modified")


class ExperimentViewer:
    @property
    def dbfile(self):
        dbfile = os.path.join(self.dbpath, self.dbfilename)
        return dbfile

    @property
    def db(self) -> PickleDB:
        return self.__db

    def __init__(self, dbfilename=None, dbpath=None):
        if dbfilename is None:
            dbfilename = "thexp.json"
        self.dbfilename = dbfilename
        if dbpath is None:
            dbpath = gu.home_dir()
        self.dbpath = dbpath

        self.__db = PickleDB(self.dbfile, auto_dump=False)

    def reload(self):
        self.__db = PickleDB(self.dbfile, auto_dump=False)

    def list(self):
        return list(self.__db.getall())

    def recent(self, day=0, mouth=0, year=0, timedeltaobj=None):
        """
        mouth和year不精确，mouth = 30*day, year = 365*day
        :param day:
        :param mouth:
        :param year:
        :return:
        """
        if timedeltaobj is not None:
            history = datetime.now() - timedeltaobj
        else:
            history = datetime.now() - timedelta(day + mouth * 30 + year * 365)
        return [date for date in self.list() if gu.curent_date(dateobj=history) < date]

    def drop(self):
        self.__db.deldb()

    def backtracking(self, fn):
        checkhash = os.path.exists(fn)
        if checkhash:
            tfhash = gu.file_atime_hash(fn)
        path = tree()

        for dkey in self.__db.getall():
            exp = self.__db.get(dkey)  # type:dict

            for dk, dv in exp.items():
                if dk == "files":
                    for k, v in dv.items():  # type:(str,Any)
                        if gu.path_in(fn, k):
                            if checkhash:
                                fhash = v["hash"]
                                if fhash != tfhash:
                                    path[dkey][dk][k] = self.check_file_state(fn, v)
                                else:
                                    path[dkey][dk][k] = v
                            else:
                                path[dkey][dk][k] = v
                elif dk == "snap":
                    for k, v in dv.items():
                        if gu.path_in(fn, v["fn"]) or gu.path_in(fn, v["snap"]):
                            path[dkey][dk] = dv
        return path

    def check_snap_state(self, file, oldstat):
        res = dict()
        if not os.path.exists(file):
            res["mode"] = "deleted"
        else:
            nhash = gu.file_atime_hash(file)
            if oldstat["hash"] != nhash:
                res["mode"] = "replacement"
                res["atime"] = oldstat["atime"]
        return res

    def check_file_state(self, file, oldstat):
        res = dict()
        if not os.path.exists(file):
            res["mode"] = "deleted"
        else:
            nhash = gu.file_atime_hash(file)
            if oldstat["hash"] != nhash:
                res["mode"] = "replacement"
                res["atime"] = oldstat["atime"]
        return res

    def exp_info(self, datekey):
        return self.__db.get(datekey)

    def exp_code(self, datekey):
        return self.__db.get(datekey)["snap"]


"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""
import pprint as pp
from collections import namedtuple

from tensorboard.backend.event_processing import event_accumulator
#
# ScalarEvent = namedtuple('ScalarEvent', ['wall_time', 'step', 'value'])
# CompressedHistogramEvent = namedtuple('CompressedHistogramEvent',
#                                       ['wall_time', 'step',
#                                        'compressed_histogram_values'])
# HistogramEvent = namedtuple('HistogramEvent',
#                             ['wall_time', 'step', 'histogram_value'])
# HistogramValue = namedtuple('HistogramValue', ['min', 'max', 'num', 'sum',
#                                                'sum_squares', 'bucket_limit',
#                                                'bucket'])
# ImageEvent = namedtuple('ImageEvent', ['wall_time', 'step',
#                                        'encoded_image_string', 'width',
#                                        'height'])
# AudioEvent = namedtuple('AudioEvent', ['wall_time', 'step',
#                                        'encoded_audio_string', 'content_type',
#                                        'sample_rate', 'length_frames'])
# TensorEvent = namedtuple('TensorEvent', ['wall_time', 'step', 'tensor_proto'])


class Plotter:
    """Extract image/scalars/... from tensorboard files and save it to images"""

    def __init__(self, plot_dir):
        self.plot_dir = plot_dir

    def dynamic_board(self, file, mode=None):
        if mode is None or mode == "created":
            self.ea = event_accumulator.EventAccumulator(file,
                                                         size_guidance={  # see below regarding this argument
                                                             event_accumulator.COMPRESSED_HISTOGRAMS: 500,
                                                             event_accumulator.IMAGES: 4,
                                                             event_accumulator.AUDIO: 4,
                                                             event_accumulator.SCALARS: 0,
                                                             event_accumulator.HISTOGRAMS: 1,
                                                         })
            print("\nLoad board in {}".format(file))

    def get_scalars(self, tag):
        return self.ea.Scalars(tag)  # 'wall_time', 'step', 'value'

    def get_histograms(self, tag):
        return self.ea.Histograms(tag)

    def get_audio(self, tag):
        return self.ea.Audio(tag)

    def get_images(self, tag):
        return self.ea.Images(tag)

    def get_compressed_histograms(self, tag):
        return self.ea.CompressedHistograms(tag)

    def get_tensor(self, tag):
        return self.ea.Tensors(tag)

    def summary(self):
        self.ea.Reload()
        pp.pprint(self.ea.Tags())



"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""

import torch

def split_sub_matrix(mat: torch.Tensor, *sizes):
    """
    将一个[N,M,...,L]的矩阵按 n,m,...l 拆分成 N/n*M/m*...L/l 个 [n,m,...l]的小矩阵

    如果N/n 无法整除，不会报错而是会将多余的裁掉
    example:
        mat = torch.arange(0,24).view(4,6) # shape = [4, 6]
        >> tensor([[ 0,  1,  2,  3,  4,  5],
                   [ 6,  7,  8,  9, 10, 11],
                   [12, 13, 14, 15, 16, 17],
                   [18, 19, 20, 21, 22, 23]])

        split_sub_matrix(mat,2,3) # shape = [2, 2, 2, 3]
        >> tensor([[[[ 0,  1,  2],
                      [ 6,  7,  8]],

                     [[ 3,  4,  5],
                      [ 9, 10, 11]]],

                    [[[12, 13, 14],
                      [18, 19, 20]],

                     [[15, 16, 17],
                      [21, 22, 23]]]])

    :param mat: 一个[N,M,...L] 的矩阵
    :param sizes: n,m,...l 的list, 其长度不一定完全和mat的维数相同
        mat = torch.arange(0,240).view([4,6,10])
        split_sub_matrix(mat,2,3) # shape = [2, 2, 10, 2, 3]
    :return: 一个 [N/row,M/col,row,col] 的矩阵
    """
    for i,size in enumerate(sizes):
        mat = mat.unfold(i,size,size)
    return mat


def cartesian_product(left: torch.Tensor, right: torch.Tensor):
    """
    笛卡尔积
    example:
        cartesian_product(torch.arange(0,3),torch.arange(0,5))
        >> (tensor([0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2]),
            tensor([0, 1, 2, 3, 4, 0, 1, 2, 3, 4, 0, 1, 2, 3, 4]))
    :param mat:
    :return: 两个矩阵
    """
    # za = mat.repeat_interleave(mat.shape[0], dim=0)
    # zb = mat.repeat([mat.shape[0], 1])
    nleft = left.repeat_interleave(right.shape[0], dim=0)
    nright = right.repeat(*[item if i == 0 else 1 for i, item in enumerate(left.shape)])
    return nleft, nright


"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""

from typing import Any

class Merge(type):
    """
    元类，用于将子类和父类共有字典，集合时，子类的覆盖行为改为合并父类的字典，集合

    由于用途特殊，仅识别类变量中以下划线开头的变量
    ::
        class A(metaclass=Merge):
            _dicts = {"1": 2, "3": 4}

        class B(A):
            _dicts = {"5":6,7:8}

        print(B._dicts)

    result:
    >>> {'5': 6, '3': 4, '1': 2, 7: 8}
    """

    def __new__(cls, name, bases, attrs: dict, **kwds):
        for base in bases:
            for key, value in base.__dict__.items():  # type:(str,Any)
                if not key.startswith("_"):
                    continue
                if isinstance(value, set):
                    v = attrs.setdefault(key, set())
                    v.update(value)
                elif isinstance(value, dict):
                    v = attrs.setdefault(key, dict())
                    v.update(value)

        return type.__new__(cls, name, bases, dict(attrs))




"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""
import collections
import shutil
import sys
import time

import numpy as np


def get_consolo_width():
    return shutil.get_terminal_size().columns


class ScreenStr():
    """
    该方法用于长期输出在同一行（如batch级别的loss输出）时，控制屏幕输出位于同一行，支持中英文混合
    该方法效率比较低，需要经过一次调用系统命令，两次对文本的编码解码和（最多）三次嵌套异常处理，
    因此可能适用场景也就限于炼丹了吧（笑
    """
    t = 0
    dt = 0.7
    last = 0
    left = 0
    max_wait = 1.
    wait = 0
    wait_toggle = False

    debug = False
    last_width = 0

    def __init__(self, content="", leftoffset=0) -> None:
        self.content = content
        ScreenStr.left = leftoffset

    def __repr__(self) -> str:
        return self._screen_str()

    def tostr(self):
        return self.content

    @staticmethod
    def set_speed(dt: float = 0.05):
        ScreenStr.dt = dt

    @staticmethod
    def deltatime():
        if ScreenStr.last == 0:
            ScreenStr.last = time.time()
            return 0
        else:
            end = time.time()
            res = end - ScreenStr.last
            ScreenStr.last = end
            return res

    @staticmethod
    def cacu_offset_(out_width):

        delta = ScreenStr.deltatime()
        ScreenStr.t += delta * ScreenStr.dt

        # pi = 2*math.pi
        t = ScreenStr.t
        # k = 2 * out_width / pi
        k = 10
        pi = 2 * out_width / k
        offset = round(k * (t % pi) * ((t % pi) < pi / 2) + (-k * (t % pi) + 2 * out_width) * ((t % pi) > pi / 2))

        # offset = math.floor(out_width * (math.cos(ScreenStr.t + math.pi) + 1) / 2)
        # print(offset)
        return offset

    a = 1

    @staticmethod
    def cacu_offset(h):
        """_/-\_"""
        delta = ScreenStr.deltatime()
        ScreenStr.t += delta * ScreenStr.dt

        t = ScreenStr.t
        k = 10
        a = ScreenStr.a
        b = h / k
        period = 2 * (a + b)

        offset = ((t % period) - a) * k * (a <= (t % period) < (a + b)) + \
                 h * ((a + b) <= (t % period) < (2 * a + b)) + \
                 (h - ((t % period) - 2 * a - b) * k) * ((2 * a + b) <= (t % period))
        offset = round(offset)
        return offset

    def __len__(self) -> int:
        txt = self.content.encode("gbk")
        return len(txt)

    def _decode_sub(self, txt, left, right):
        try:
            txt = txt[left:right].decode("gbk")
        except:
            try:
                txt = txt[left:right - 1].decode("gbk")
            except:
                try:
                    txt = txt[left + 1:right].decode("gbk")
                except:
                    txt = txt[left + 1:right - 1].decode("gbk")

        return txt

    @staticmethod
    def refresh():
        ScreenStr.t = 0
        ScreenStr.dt = abs(ScreenStr.dt)
        ScreenStr.last = 0

    @staticmethod
    def consolo_width():
        width = get_consolo_width()
        return width

    @staticmethod
    def split(txt, len):
        try:
            return txt[:len], txt[len:]
        except:
            try:
                return txt[:len + 1], txt[len + 1:]
            except:
                return txt[:len - 1], txt[len - 1:]

    def _screen_str(self, margin="..."):
        width = ScreenStr.consolo_width()

        txt = self.content.encode("gbk").strip()
        textlen = len(txt)

        if textlen <= width:
            return self.content

        left, right = ScreenStr.split(txt, ScreenStr.left)
        if len(left) >= width:
            return left[:width]

        offset = ScreenStr.cacu_offset(len(right) - (width) + 4)

        offright = width + offset - len(margin)

        left = left.decode("gbk")
        right = self._decode_sub(right, offset, offright)

        head = "\r" if self.content.startswith("\r") else ""
        tail = "\n" if self.content.endswith("\n") else ""

        txt = "{}{}{}{}".format(head, left, right, tail)
        return txt + margin


class Progbar(object):
    """Displays a progress bar.

    # Arguments
        target: Total number of steps expected, None if unknown.
        width: Progress bar width on screen.
        verbose: Verbosity mode, 0 (silent), 1 (verbose), 2 (semi-verbose)
        stateful_metrics: Iterable of string names of metrics that
            should *not* be averaged over time. Metrics in this list
            will be displayed as-is. All others will be averaged
            by the progbar before display.
        interval: Minimum visual progress update interval (in seconds).
    """

    def __init__(self, target, width=30, verbose=1, interval=0.05,
                 stateful_metrics=None):
        self.target = target
        self.width = width
        self.verbose = verbose
        self.interval = interval
        if stateful_metrics:
            self.stateful_metrics = set(stateful_metrics)
        else:
            self.stateful_metrics = set()

        self._dynamic_display = ((hasattr(sys.stdout, 'isatty') and
                                  sys.stdout.isatty()) or
                                 'ipykernel' in sys.modules)
        self._total_width = 0
        self._seen_so_far = 0
        self._values = collections.OrderedDict()
        self._start = time.time()
        self._last_update = 0

    def update(self, current, values=None):
        """Updates the progress bar.

        # Arguments
            current: Index of current step.
            values: List of tuples:
                `(name, value_for_last_step)`.
                If `name` is in `stateful_metrics`,
                `value_for_last_step` will be displayed as-is.
                Else, an average of the metric over time will be displayed.
        """
        strs = []
        values = values or []
        for k, v in values:
            if k not in self.stateful_metrics:
                if k not in self._values:
                    self._values[k] = [v * (current - self._seen_so_far),
                                       current - self._seen_so_far]
                else:
                    self._values[k][0] += v * (current - self._seen_so_far)
                    self._values[k][1] += (current - self._seen_so_far)
            else:
                # Stateful metrics output a numeric value.  This representation
                # means "take an average from a single value" but keeps the
                # numeric formatting.
                self._values[k] = [v, 1]
        self._seen_so_far = current

        now = time.time()
        info = ' - %.0fs' % (now - self._start)
        if self.verbose == 1:
            if (now - self._last_update < self.interval and
                    self.target is not None and current < self.target):
                return

            prev_total_width = self._total_width

            if self.target is not None:
                numdigits = int(np.floor(np.log10(self.target))) + 1
                barstr = '%%%dd/%d [' % (numdigits, self.target)
                bar = barstr % current
                prog = float(current) / self.target
                prog_width = int(self.width * prog)
                if prog_width > 0:
                    bar += ('=' * (prog_width - 1))
                    if current < self.target:
                        bar += '>'
                    else:
                        bar += '='
                bar += ('.' * (self.width - prog_width))
                bar += ']'
            else:
                bar = '%7d/Unknown' % current

            self._total_width = len(bar)
            strs.append(bar)

            if current:
                time_per_unit = (now - self._start) / current
            else:
                time_per_unit = 0
            if self.target is not None and current < self.target:
                eta = time_per_unit * (self.target - current)
                if eta > 3600:
                    eta_format = ('%d:%02d:%02d' %
                                  (eta // 3600, (eta % 3600) // 60, eta % 60))
                elif eta > 60:
                    eta_format = '%d:%02d' % (eta // 60, eta % 60)
                else:
                    eta_format = '%ds' % eta

                info = ' - ETA: %s' % eta_format
            else:
                if time_per_unit >= 1:
                    info += ' %.0fs/step' % time_per_unit
                elif time_per_unit >= 1e-3:
                    info += ' %.0fms/step' % (time_per_unit * 1e3)
                else:
                    info += ' %.0fus/step' % (time_per_unit * 1e6)

            for k in self._values:
                info += ' - %s:' % k
                if isinstance(self._values[k], list):
                    avg = np.mean(
                        self._values[k][0] / max(1, self._values[k][1]))
                    if abs(avg) > 1e-3:
                        info += ' %.4f' % avg
                    else:
                        info += ' %.4e' % avg
                else:
                    info += ' %s' % self._values[k]

            self._total_width += len(info)
            if prev_total_width > self._total_width:
                info += (' ' * (prev_total_width - self._total_width))

            if self.target is not None and current >= self.target:
                info += '\n'

            strs.append(info)

        elif self.verbose == 2:
            if self.target is None or current >= self.target:
                for k in self._values:
                    info += ' - %s:' % k
                    avg = np.mean(
                        self._values[k][0] / max(1, self._values[k][1]))
                    if avg > 1e-3:
                        info += ' %.4f' % avg
                    else:
                        info += ' %.4e' % avg
                info += '\n'

                strs.append(info)

        self._last_update = now

        return "".join(strs)

    def add(self, n, values=None):
        self.update(self._seen_so_far + n, values)


if __name__ == '__main__':
    for i in range(100):
        print("\r" + str(ScreenStr(str([i for i in range(30)]), leftoffset=10)), end="")
        time.sleep(0.2)



"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""
from collections import OrderedDict
from typing import Any

import torch

from ..base_classes.trickitems import AvgItem, NoneItem


class Meter:
    """
    m = Meter()
    m.short()
    m.int()
    m.float()
    m.percent()
    ...

    m.k += 10
    m.k.update()

    m.k.backstep()

    p = Param()
        list/tensor/optim
    """

    def __init__(self):
        self._param_dict = OrderedDict()
        self._format_dict = dict()
        self._convert_type = []

    def int(self, item):
        self._format_dict[item] = lambda x: "{:.0f}".format(x)

    def float(self, item, acc=4):
        self._format_dict[item] = lambda x: "{{:.{}f}}".format(acc).format(x)

    def percent(self, item, acc=2):
        self._format_dict[item] = lambda x: "{{:.{}%}}".format(acc).format(x)

    def tensorfloat(self, item, acc=4):
        def func(x: torch.Tensor):
            if len(x.shape) == 0:
                return "{{:.{}f}}".format(acc).format(x)
            else:
                return "{{:.{}f}}".format(acc).format(x.item())

        self._format_dict[item] = func

    def str_in_line(self, item):
        def func(x):
            l = str(x).split("\n")
            if len(l) > 1:
                return "{}...".format(l[0])
            else:
                return l[0]

        self._format_dict[item] = func

    def add_format_type(self, type, func):
        self._convert_type.append((type, func))

    def _convert(self, val):
        if type(val) in {int, float, bool, str}:
            return val
        # elif isinstance(val, torch.Tensor):
        #     if len(val.shape) == 1 and val.shape[0] == 1:
        #         return val[0]
        for tp, func in self._convert_type:
            if type(val) == tp:
                val = func(val)
        return val

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            super().__setattr__(name, value)
        elif name.endswith("_"):
            assert False
        else:
            value = self._convert(value)
            if isinstance(value, int) and name not in self._format_dict:
                self.int(name)
            elif isinstance(value, float) and name not in self._format_dict:
                self.float(name)
            elif isinstance(value, torch.Tensor) and name not in self._format_dict:
                if len(value.shape) == 0 or (sum(value.shape) == 1):
                    self.tensorfloat(name)
                else:
                    self.str_in_line(name)

            self._param_dict[name] = value

    def __setitem__(self, key, value):
        key = str(key)
        self.__setattr__(key, value)

    def __getattr__(self, item):
        if item.endswith("_"):
            return item.rstrip("_")

        if item not in self._param_dict:
            return NoneItem()
        else:
            return self._param_dict[item]

    def serialize(self):
        log_dict = OrderedDict()
        for k, v in self._param_dict.items():
            if k in self._format_dict:
                v = self._format_dict[k](v)
            log_dict[k] = v
        return log_dict

    def __getitem__(self, item):
        item = str(item)
        return self.__getattr__(item)

    def __repr__(self):
        log_dict = self.serialize()

        return "  ".join(["@{}={}".format(k, v) for k, v in log_dict.items()])


class AvgMeter(Meter):
    def __init__(self):
        super().__init__()

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            super().__setattr__(name, value)
        elif name.endswith("_"):
            assert False
        else:
            value = self._convert(value)
            if isinstance(value, int) and name not in self._format_dict:
                self.int(name)
            elif isinstance(value, float) and name not in self._format_dict:
                self.float(name)

            if isinstance(value, (float)):
                if name not in self._param_dict:
                    self._param_dict[name] = AvgItem()
                self._param_dict[name].update(value)
            elif isinstance(value, torch.Tensor):
                value = value.detach().cpu().numpy()
                if len(value.shape) == 0 or sum(value.shape) == 1:
                    if name not in self._param_dict:
                        self._param_dict[name] = AvgItem()
                    self._param_dict[name].update(value)
                else:
                    self._param_dict[name] = value
            else:
                self._param_dict[name] = value

    def update(self, meter):
        for k, v in meter._param_dict.items():
            self[k] = v
        self._format_dict.update(meter._format_dict)
        self._convert_type.extend(meter._convert_type)

    def serialize(self):
        log_dict = OrderedDict()
        for k, v in self._param_dict.items():
            if k in self._format_dict:
                v = self._format_dict[k](v)
            log_dict[k] = v
        return log_dict


"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""


import numpy as np


def sigmoid_rampup(current, rampup_length=100, amplitude=1):
    """Exponential rampup from https://arxiv.org/abs/1610.02242"""

    if rampup_length == 0:
        return 1.0
    else:
        current = np.clip(current, 0.0, rampup_length)
        phase = 1.0 - current / rampup_length
        return float(np.exp(-5.0 * phase * phase))


def linear_rampup(current, rampup_length=100, amplitude=1, offset=0):
    """
    y
    \  current
    \    ↓   ↓ rampup_length
    \   ↓   -----------   ← amplitude
    \  ↓ /
    \ /
    \ }   offset (independent of amplitude)
    \-----------------------> x
    \
    \
    :param current:
    :param rampup_length:
    :param amplitude:
    :param offset:
    :return:
    """


    """Linear rampup"""
    percent = current / rampup_length
    if percent > 1:
        return amplitude
    else:
        return percent * amplitude


def cosine_rampdown(current, rampdown_length):
    """Cosine rampdown from https://arxiv.org/abs/1608.03983"""
    assert 0 <= current <= rampdown_length
    # if current > rampdown_length:
    #     return float(.5 * (np.cos(np.pi) + 1))
    return float(.5 * (np.cos(np.pi * current / rampdown_length) + 1))



"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""
import os
from datetime import datetime
from typing import Any

from ..utils.generel_util import curent_date
from ..utils.screen import ScreenStr


class Logger:
    VERBOSE = 0
    V_DEBUG = -1
    V_INFO = 0
    V_WARN = 1
    V_ERROR = 2
    V_FATAL = 3
    _instance = None

    def __new__(cls, *args, **kwargs) -> Any:
        if Logger._instance is not None:
            return Logger._instance
        return super().__new__(cls)

    def __init__(self, adddate=True, datefmt: str = '%y-%m-%d %H:%M:%S', sep: str = " - ", stdout=True):
        """
        :param datefmt:
        :param sep:
        """
        if Logger._instance is not None:
            return

        self.adddate = adddate
        self.datefmt = datefmt
        self.out_channel = []
        self.pipe_key = set()
        self.sep = sep
        self.return_str = ""
        self.listener = []
        self.stdout = stdout
        Logger._instance = self

    def format(self, *values, prefix="", inline=False, fix=0):
        """根据初始化设置 格式化 前缀和LogMeter"""
        if self.adddate:
            cur_date = datetime.now().strftime(self.datefmt)
        else:
            cur_date = ""

        space = [cur_date, "{}".format(prefix), *["{}".format(str(i)) for i in values]]
        space = [i for i in space if len(i.strip()) != 0]

        if fix >= 0:
            left, right = self.sep.join(space[:fix + 1]), self.sep.join(space[fix + 1:])
            fix = len(left)
            logstr = self.sep.join((left, right))

            if inline:
                return "\r{}".format(logstr), fix
            else:
                return "{}\n".format(logstr), fix

        space = self.sep.join(space)

        if inline:
            return "\r{}".format(space), 0
        else:
            return "{}\n".format(space), 0

    def inline(self, *values, prefix="", fix=0):
        """在一行内输出 前缀 和 LogMeter"""
        logstr, fix = self.format(*values, prefix=prefix, inline=True, fix=fix)

        self.handle(logstr, fix=fix)

    def info(self, *values, prefix=""):
        """以行为单位输出 前缀 和 LogMeter"""
        logstr, fix = self.format(*values, prefix=prefix, inline=False)
        self.handle(logstr, level=Logger.V_INFO, fix=fix)

    def debug(self, *values, prefix=""):
        logstr, fix = self.format("DEBUG", *values, prefix=prefix, inline=False)
        self.handle(logstr, level=Logger.V_DEBUG, fix=fix)

    def warn(self, *values, prefix=""):
        logstr, fix = self.format("WARN", *values, prefix=prefix, inline=False)
        self.handle(logstr, level=Logger.V_WARN, fix=fix)

    def error(self, *values, prefix=""):
        logstr, fix = self.format("ERROR", *values, prefix=prefix, inline=False)
        self.handle(logstr, level=Logger.V_ERROR, fix=fix)

    def fatal(self, *values, prefix=""):
        logstr, fix = self.format("FATAL", *values, prefix=prefix, inline=False)
        self.handle(logstr, level=Logger.V_FATAL, fix=fix)

    def newline(self):
        """换行"""
        self.handle("\n")

    def handle(self, logstr, end="", level=0, **kwargs):
        """
        handle log stinrg，以指定的方式输出
        :param logstr:
        :param _:
        :param end:
        :return:
        """
        for listener in self.listener:
            listener(logstr, end, level)

        if level < Logger.VERBOSE:
            return

        if logstr.startswith("\r"):
            fix = kwargs.get("fix", 0)
            self.return_str = logstr
            self.print(ScreenStr(logstr, leftoffset=fix), end=end)
        else:
            if len(self.return_str) != 0:
                self.print(self.return_str, end="\n")
            self.print(logstr, end="")

            for i in self.out_channel:
                with open(i, "a", encoding="utf-8") as w:
                    if len(self.return_str) != 0:
                        w.write("{}\n".format(self.return_str.strip()))
                    w.write(logstr)

            self.return_str = ""

    def print(self, *args, end='\n'):
        if self.stdout:
            print(*args, end=end, flush=True)

    def enbale_stdout(self, val):
        self.stdout = val

    def add_log_dir(self, dir):
        """添加一个输出到文件的管道"""
        if dir in self.pipe_key:
            self.info("Add pipe {}, but already exists".format(dir))
            return False

        os.makedirs(dir, exist_ok=True)

        i = 0
        cur_date = curent_date(fmt="%y%m%d%H%M%S")
        fni = os.path.join(dir, "{}.{}.log".format(cur_date, i))
        while os.path.exists(fni):
            i += 1
            fni = os.path.join(dir, "{}.{}.log".format(cur_date, i))

        self.print("add output channel on {}".format(fni))
        self.out_channel.append(fni)
        self.pipe_key.add(dir)
        return True

    def add_log_listener(self, func):
        self.listener.append(func)



"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""
import hashlib
import os
import sys
from datetime import datetime


def listdir_by_time(dir_path):
    dir_list = os.listdir(dir_path)
    if not dir_list:
        return []
    else:
        # os.path.getctime() 函数是获取文件最后创建时间
        dir_list = sorted(dir_list, key=lambda x: os.path.getatime(os.path.join(dir_path, x)), reverse=True)
        # print(dir_list)

        # for i in dir_list:
        #     print(os.path.getctime(os.path.join(dir_path,i)),i)
        return dir_list


def home_dir():
    path = os.path.expanduser("~/.thexp")
    os.makedirs(path, exist_ok=True)
    return path


def file_atime_hash(file):
    return string_hash(str(os.path.getatime(file)))


def string_hash(*str):
    hl = hashlib.md5()
    for s in str:
        hl.update(s.encode(encoding='utf-8'))
    return hl.hexdigest()[:16]


def file_hash(file):
    hl = hashlib.md5()
    with open(file, encoding="utf-8") as r:
        s = "".join(r.readlines())
        hl.update(s.encode(encoding='utf-8'))
    return hl.hexdigest()[:16]


def curent_date(fmt='%y-%m-%d-%H%M%S', dateobj: datetime = None):
    if dateobj is not None:
        return dateobj.strftime(fmt)
    return datetime.now().strftime(fmt)


def file_atime2date(file, fmt='%y%m%d-%H%M%S'):
    return curent_date(fmt, datetime.fromtimestamp(os.path.getatime(file)))


def path_equal(p1, p2):
    return os.path.normcase(p1) == os.path.normcase(p2)


def path_in(sub, all):
    return os.path.normcase(sub) in os.path.normcase(all)


class exithook():
    def __init__(self):
        self.exit_code = None
        self.exception = None

    def hook(self):
        self._orig_exit = sys.exit
        sys.exit = self.exit
        sys.excepthook = self.exc_handler

    def exit(self, code=0):
        self.exit_code = code
        self._orig_exit(code)

    def exc_handler(self, exc_type, exc, *args):
        self.exception = exc



"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""

import os
from functools import wraps

from .meter import AvgMeter
from .meter import Meter
from .params import Params
from .trainer import BaseTrainer


class BaseCallback():
    """
    基类

    除了基本的回调接口外，还实现了`auto_hook`和`reverse_hook`两个方法，
    用来将callback实现的方法绑定到trainer和将trainer所有可回调的函数绑定到callback中。

    因为绑定的方法只会调用on_begin()和on_end()，因此对于具体的方法需要进行判断进行方法的分流，或者不使用自动绑定，
    而是主动用trainer绑定::

        trainer.add_callback(func=trainer.train, callback=cb)

    """
    priority = 0  # 所有内部实现的Callback优先级均在 [0-100] 以内

    def __new__(cls, *_, **__):
        self = super().__new__(cls)
        self.enable = True

        def ecp_wrap(func):
            """同一个异常第一次调用的时候运行"""

            @wraps(func)
            def on_exception(trainer: BaseTrainer, tfunc, param: Params, e: BaseException, *args, **kwargs):
                self.ecp = getattr(self, "ecp", None)
                if self.ecp != e:
                    self.on_first_exception(trainer, tfunc, param, e, *args, **kwargs)
                    self.ecp = e
                func(trainer, tfunc, param, e, *args, **kwargs)

            return on_exception

        self.on_exception = ecp_wrap(self.on_exception)
        return self

    def on_hooked(self, trainer: BaseTrainer, param: Params):
        """called when callback hooked trainer"""
        trainer.logger.info("{} hooked on {}.".format(self, trainer))

    def on_first_exception(self, trainer: BaseTrainer, func, param: Params, e: BaseException, *args, **kwargs):
        """
        when an exception was raised in some function, on_exception() will be called.

        如果异常发生在一个嵌套调用的函数中，那么该异常会在每一层函数都raise一次。

        该方法将被调用当该异常第一次raise出来的时候。
        该方法在 __new__ 中作了处理逻辑，不受继承关系影响
        """
        pass

    def on_exception(self, trainer: BaseTrainer, func, param: Params, e: BaseException, *args, **kwargs):
        """called when exception raised in some function"""
        return False

    def on_hook_failed(self, trainer, message):
        """Any reason when callback cannot hook on trainer"""
        pass

    def on_begin(self, trainer: BaseTrainer, func, param: Params, *args, **kwargs):
        """called before trainer.func is called"""
        pass

    def on_end(self, trainer: BaseTrainer, func, param: Params, meter, *args, **kwargs):
        pass

    def __le__(self, other):
        return self.priority <= other.priority

    def __lt__(self, other):
        return self.priority < other.priority

    def hook(self, trainer: BaseTrainer):
        """自动将自己已有的on_func_begin/on_func_end方法绑定"""
        trainer.add_callback(self)

    def unhook(self):
        self._trainer.remove_callback(self)

    def _repr_by_val(self, *vals):
        vstr = "; ".join(["{}={}".format(val, str(getattr(self, val, None))) for val in vals])
        return "{}([{}])".format(self.__class__.__name__, vstr)

    def __repr__(self) -> str:
        return self._repr_by_val("priority")


class TrainCallback(BaseCallback):
    """
    实现了一般训练过程中的函数函数回调
    """

    def on_begin(self, trainer: BaseTrainer, func, param: Params, *args, **kwargs):
        if func.__name__ == "train":
            self.on_train_begin(trainer, func, param, *args, **kwargs)
        elif func.__name__ == "train_epoch":
            self.on_train_epoch_begin(trainer, func, param, *args, **kwargs)
        elif func.__name__ == "train_batch":
            self.on_train_batch_begin(trainer, func, param, *args, **kwargs)
        elif func.__name__ == "tests":
            self.on_test_begin(trainer, func, param, *args, **kwargs)
        elif func.__name__ == "eval":
            self.on_eval_begin(trainer, func, param, *args, **kwargs)

    def on_train_begin(self, trainer: BaseTrainer, func, param: Params, *args, **kwargs):
        pass

    def on_train_epoch_begin(self, trainer: BaseTrainer, func, param: Params, *args, **kwargs):
        pass

    def on_test_begin(self, trainer: BaseTrainer, func, param: Params, *args, **kwargs):
        pass

    def on_eval_begin(self, trainer: BaseTrainer, func, param: Params, *args, **kwargs):
        pass

    def on_train_batch_begin(self, trainer: BaseTrainer, func, param: Params, *args, **kwargs):
        pass

    def on_end(self, trainer: BaseTrainer, func, param: Params, meter, *args, **kwargs):
        if func.__name__ == "train":
            self.on_train_end(trainer, func, param, meter, *args, **kwargs)
        elif func.__name__ == "train_epoch":
            self.on_train_epoch_end(trainer, func, param, meter, *args, **kwargs)
        elif func.__name__ == "train_batch":
            self.on_train_batch_end(trainer, func, param, meter, *args, **kwargs)
        elif func.__name__ == "tests":
            self.on_test_end(trainer, func, param, meter, *args, **kwargs)
        elif func.__name__ == "eval":
            self.on_eval_end(trainer, func, param, meter, *args, **kwargs)

    def on_train_end(self, trainer: BaseTrainer, func, param: Params, meter: Meter, *args, **kwargs):
        pass

    def on_train_epoch_end(self, trainer: BaseTrainer, func, param: Params, meter: Meter, *args, **kwargs):
        pass

    def on_test_end(self, trainer: BaseTrainer, func, param: Params, meter: Meter, *args, **kwargs):
        pass

    def on_eval_end(self, trainer: BaseTrainer, func, param: Params, meter: Meter, *args, **kwargs):
        pass

    def on_train_batch_end(self, trainer: BaseTrainer, func, param: Params, meter: Meter, *args, **kwargs):
        pass


class EvalCallback(TrainCallback):
    """决定在训练过程中，几个epoch eval一次，几个epoch  test一次"""

    def __init__(self, eval_per_epoch=1, test_per_epoch=10):
        self.eval_in_per_epoch = eval_per_epoch
        self.test_in_per_epoch = test_per_epoch

    def on_train_epoch_end(self, trainer: BaseTrainer, func, param: Params, meter: Meter, *args, **kwargs):
        if param.eidx % self.eval_in_per_epoch == self.eval_in_per_epoch - 1:
            trainer.eval()
        if param.eidx % self.test_in_per_epoch == self.test_in_per_epoch - 1:
            trainer.test()

    def on_train_end(self, trainer: BaseTrainer, func, param: Params, meter: Meter, *args, **kwargs):
        if param.eidx % self.eval_in_per_epoch != self.eval_in_per_epoch - 1:
            trainer.eval()
        if param.eidx % self.test_in_per_epoch != self.test_in_per_epoch - 1:
            trainer.test()

    def __repr__(self):
        return self._repr_by_val("eval_in_per_epoch","test_per_epoch")


class LoggerCallback(TrainCallback):
    def __init__(self, avg=True):
        self.avg = avg

    def on_hooked(self, trainer: BaseTrainer, param: Params):
        super().on_hooked(trainer, param)
        trainer.logger.info("Exp BaseDir", os.path.abspath(trainer.experiment.exp_dir))
        trainer.logger.info("Exp Trainer", trainer.__class__.__name__)
        trainer.logger.info("Exp Params", param)

    def on_train_epoch_begin(self, trainer: BaseTrainer, func, param: Params, *args, **kwargs):
        if self.avg:
            self.meter = AvgMeter()
        trainer.logger.info("{}/{}".format(param.eidx, param.epoch), prefix="train epoch")

    def on_train_begin(self, trainer: BaseTrainer, func, param: Params, *args, **kwargs):
        super().on_train_begin(trainer, func, param, *args, **kwargs)

    def on_train_end(self, trainer: BaseTrainer, func, param: Params, meter: Meter, *args, **kwargs):
        if meter is None:
            meter = ""
        trainer.logger.info("train end", meter)

    def on_train_batch_end(self, trainer: BaseTrainer, func, param: Params, meter: Meter, *args, **kwargs):
        if meter is None:
            meter = ""
        else:

            if self.avg:
                self.meter.update(meter)
                meter = self.meter
        trainer.logger.inline("{}/{}".format(param.idx, len(trainer.iter_train_dataloader())), meter, fix=1)

    def on_first_exception(self, trainer: BaseTrainer, func, param: Params, e: BaseException, *args, **kwargs):
        trainer.logger.error("{} raised".format(e.__class__.__name__))

    def on_test_begin(self, trainer: BaseTrainer, func, param: Params, *args, **kwargs):
        trainer.logger.info("tests start")

    def on_eval_begin(self, trainer: BaseTrainer, func, param: Params, *args, **kwargs):
        trainer.logger.info("eval start")

    def on_eval_end(self, trainer: BaseTrainer, func, param: Params, meter: Meter, *args, **kwargs):
        if meter is None:
            meter = ""
        trainer.logger.info("eval end", meter)

    def on_test_end(self, trainer: BaseTrainer, func, param: Params, meter: Meter, *args, **kwargs):
        if meter is None:
            meter = ""
        trainer.logger.info("tests end", meter)




"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""

from torch.utils.data.dataloader import DataLoader as _DataLoader
from torch.utils.data.sampler import BatchSampler


class DataLoader(_DataLoader):

    def set_batch_size(self, batch_size):
        # batch_sampler = BatchSampler(self.sampler, batch_size, self.drop_last)
        # self.batch_sampler = batch_sampler
        self.batch_sampler.batch_size = batch_size
        # self.batch_size = batch_size

    # def __setattr__(self, attr, val):
        # if self.__initialized and attr in ('sampler', 'drop_last', 'dataset'):
        #     raise ValueError('{} attribute should not be set after {} is '
        #                      'initialized'.format(attr, self.__class__.__name__))
        # object.__setattr__(self,attr,val)


from setuptools import setup,find_packages
from thexp import __VERSION__
setup(
    name='thexp',
    # 主版本，次版本，修订号？，bug修订号，...待定
    version=__VERSION__,
    description='An useful torch tool for your experiment.',
    url='https://github.com/sailist/thexp',
    author='sailist',
    author_email='sailist@outlook.com',
    license='Apache License 2.0',
    include_package_data = True,
    install_requires = [
      "torch","fire","tensorboard","watchdog","docopt"
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='thexp',
    packages=find_packages(),
    entry_points={
        'console_scripts':[
            'thexp = thexp.cli.cli:main'
        ]
      },
)


"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""

from .meter import Meter,AvgMeter
from .params import Params
from .trainer import Trainer
from .saver import Saver
from .rndmanager import RndManager
from .logger import Logger
from .databundler import DataBundler
from .experiment import ExperimentViewer,Experiment


"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""

from torch import randperm
from torch._utils import _accumulate
from torch.utils.data.dataset import Subset


def random_split_with_indices(dataset, lengths,indices=None):
    r"""
    Randomly split a dataset into non-overlapping new datasets of given lengths.

    Arguments:
        dataset (Dataset): Dataset to be split
        lengths (sequence): lengths of splits to be produced
    """
    if sum(lengths) != len(dataset):
        raise ValueError("Sum of input lengths does not equal the length of the input dataset!")
    if indices is None:
        indices = randperm(sum(lengths)).tolist()

    return [Subset(dataset, indices[offset - length:offset]) for offset, length in
            zip(_accumulate(lengths), lengths)], indices



"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""

from collections import OrderedDict
from itertools import cycle, chain

from torch.utils.data import DataLoader
from ..utils.date.dataloader import DataLoader as thDataLoader
import torch


class DataBundler:
    """
    当一个模型需要训练多个数据集的时候，通过DataBundler类和附属的装饰器灵活的提供数据集::

        bundler = DataBundler() \
                .cycle(cifar_dataloader, "cifar") \
                .add(svhn_dataloader, "svhn").chain_iter()

        for (imgs,labels) in bundler:
            ...

    效果等价于::

        for (imgs,chains) in chain(cifar_dataloader,svhn_dataloader):
            ...

    另外还支持
    """

    def __init__(self):
        self.dataloaders = OrderedDict()
        self.iter_mode = "chain"

    def set_batch_size(self,batch_size):
        for _,(loader,_) in self.dataloaders.items():
            if isinstance(loader,thDataLoader):
                loader.set_batch_size(batch_size)
            elif isinstance(loader,DataLoader):
                loader.batch_sampler.batch_size = batch_size

    def _append(self, loader, func, name):
        if name is None:
            unname = "unnamed"
            i = 0
            name = "{}_{}".format(unname, i)
            while name in self.dataloaders:
                i += 1
                name = "{}_{}".format(unname, i)
        else:
            assert name not in self.dataloaders, "{} also defined in bundler".format(name)

        self.dataloaders[name] = (loader, func)

    def __len__(self):
        if self.iter_mode == 'zip':
            return max(*self.len_list())
        elif self.iter_mode == "chain":
            return sum(self.len_list())

    def len_list(self):
        """
        按照添加的顺序返回各个dataloader的长度（batch级别）
        :return:
        """
        return [len(loader) for _, (loader, _) in self.dataloaders.items()]

    def len_dict(self):
        """
        返回每个loader的 name:len 字典
        :return: an OrderedDict
        """
        res = OrderedDict()
        for name, (loader, func) in self.dataloaders.items():
            res[name] = len(loader)
        return res

    def cycle(self, loader: DataLoader, name=None):
        """一般在zip中保证数据量少的数据集不会成为拖累"""
        assert isinstance(loader, DataLoader)
        self._append(loader, cycle, name)
        return self

    def add(self, loader, name=None):
        assert isinstance(loader, DataLoader)
        self._append(loader, lambda x: x, name)
        return self

    def _func_loader(self):
        return [func(loader) for name, (loader, func) in self.dataloaders.items()]

    def __getitem__(self, item):
        return self.dataloaders[item][0]

    def __iter__(self):
        loaders = self._func_loader()
        if len(loaders) == 1:
            return iter(loaders[0])

        if self.iter_mode == "zip":
            return zip(*loaders)
        elif self.iter_mode == "chain":
            return chain(*loaders)

        assert False

    def choice_batch(self)->tuple:
        return next(iter(self))

    def choice_sample(self)->tuple:
        xs,ys = next(iter(self)) # type:(torch.Tensor,torch.Tensor)
        return (xs[0],ys[0])

    def zip_mode(self):
        """切换为zip方式提供所有添加的数据集"""
        self.iter_mode = "zip"
        return self

    def chain_mode(self):
        """
        切换为chain方式提供所有添加的数据集
            注意，如果以cycle方法添加了某个数据集，那么该迭代将永远不会停止
        :return:
        """
        self.iter_mode = "chain"
        return self

    @staticmethod
    def create_zip_bundler(**kwargs):
        bundler = DataBundler()
        loaders = [(len(v),v,k) for k,v in kwargs.items()]
        loaders.sort(reverse=True)
        bundler.add(loaders[0][1],loaders[0][2])
        for (_,loader,name) in loaders[1:]:
            bundler.cycle(loader,name)
        return bundler

    @staticmethod
    def create_chain_bundler(*args):
        bundler = DataBundler()
        for loader in args:
            bundler.add(loader)
        return bundler



"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""
import pprint as pp
from functools import partial

import torch
from torch.optim.adadelta import Adadelta
from torch.optim.adagrad import Adagrad
from torch.optim.adam import Adam
from torch.optim.adamax import Adamax
from torch.optim.adamw import AdamW
from torch.optim.asgd import ASGD
from torch.optim.lbfgs import LBFGS
from torch.optim.optimizer import Optimizer
from torch.optim.rmsprop import RMSprop
from torch.optim.rprop import Rprop
from torch.optim.sgd import SGD
from torch.optim.sparse_adam import SparseAdam


class DeviceParam:
    def __init__(self, cpu_final=False):
        devices = []
        if torch.cuda.is_available():
            for i in range(torch.cuda.device_count()):
                devices.append(torch.device("cuda:{}".format(i)))
        if cpu_final:
            devices.append(torch.device("cpu"))




class OptimParam:
    O_Adadelta = Adadelta
    O_Adagrad = Adagrad
    O_Adam = Adam
    O_AdamW = AdamW
    O_SparseAdam = SparseAdam
    O_Adamax = Adamax
    O_ASGD = ASGD
    O_LBFGS = LBFGS
    O_RMSprop = RMSprop
    O_Rprop = Rprop
    O_SGD = SGD

    def __init__(self):
        self.optim_param = dict()
        self.optim_dict = dict(
            adadelta=Adadelta,
            adagrad=Adagrad,
            adam=Adam,
            adamw=AdamW,
            sparseadam=SparseAdam,
            adamax=Adamax,
            asgd=ASGD,
            lbfgs=LBFGS,
            rmsprop=RMSprop,
            rprop=Rprop,
            sgd=SGD,
        )

    def choose(self, optimizer):
        if issubclass(optimizer, Optimizer):
            pass
        else:
            optimizer = self.optim_dict[optimizer.lower()]

    def create(self, optimizer, param):
        assert issubclass(optimizer, Optimizer)
        return optimizer(param, **self[optimizer.__name__.lower()])

    def __iter__(self):
        for k, v in self.optim_param.items():
            yield partial(self.optim_dict[k], v)

    def __setitem__(self, key, value):
        self.optim_param[key] = value

    def __getitem__(self, item):
        return self.optim_param[item]

    class OptimBuilder():
        def __init__(self, param):
            self.param = param  # type:(OptimParam)

        def finish(self):
            return self.param

        def Adadelta(self, lr=1.0, rho=0.9, eps=1e-06, weight_decay=0):
            self.param["adadelta"] = dict(
                lr=lr,
                rho=rho,
                eps=eps,
                weight_decay=weight_decay,
            )
            return self

        def Adagrad(self, lr=0.01, lr_decay=0, weight_decay=0, initial_accumulator_value=0, eps=1e-10):
            self.param["adagrad"] = dict(
                lr=lr,
                lr_decay=lr_decay,
                weight_decay=weight_decay,
                initial_accumulator_value=initial_accumulator_value,
                eps=eps,
            )
            return self

        def Adam(self, lr=0.001, betas=(0.9, 0.999), eps=1e-08, weight_decay=0, amsgrad=False):
            self.param["adam"] = dict(
                lr=lr,
                betas=betas,
                eps=eps,
                weight_decay=weight_decay,
                amsgrad=amsgrad,
            )
            return self

        def AdamW(self, lr=0.001, betas=(0.9, 0.999), eps=1e-08, weight_decay=0.01, amsgrad=False):
            self.param["adamw"] = dict(
                lr=lr,
                betas=betas,
                eps=eps,
                weight_decay=weight_decay,
                amsgrad=amsgrad,
            )
            return self

        def SparseAdam(self, lr=0.001, betas=(0.9, 0.999), eps=1e-08):
            self.param["sparseadam"] = dict(
                lr=lr,
                betas=betas,
                eps=eps,
            )
            return self

        def Adamax(self, lr=0.002, betas=(0.9, 0.999), eps=1e-08, weight_decay=0):
            self.param["adamax"] = dict(
                lr=lr,
                betas=betas,
                eps=eps,
                weight_decay=weight_decay,
            )
            return self

        def ASGD(self, lr=0.01, lambd=0.0001, alpha=0.75, t0=1000000.0, weight_decay=0):
            self.param["asgd"] = dict(
                lr=lr,
                lambd=lambd,
                alpha=alpha,
                t0=t0,
                weight_decay=weight_decay,
            )
            return self

        def LBFGS(self, lr=1, max_iter=20, max_eval=None, tolerance_grad=1e-07, tolerance_change=1e-09,
                  history_size=100, line_search_fn=None):
            self.param["lbfgs"] = dict(
                lr=lr,
                max_iter=max_iter,
                max_eval=max_eval,
                tolerance_grad=tolerance_grad,
                tolerance_change=tolerance_change,
                history_size=history_size,
                line_search_fn=line_search_fn,
            )
            return self

        def RMSprop(self, lr=0.01, alpha=0.99, eps=1e-08, weight_decay=0, momentum=0, centered=False):
            self.param["rmsprop"] = dict(
                lr=lr,
                alpha=alpha,
                eps=eps,
                weight_decay=weight_decay,
                momentum=momentum,
                centered=centered,
            )
            return self

        def Rprop(self, lr=0.01, etas=(0.5, 1.2), step_sizes=(1e-06, 50)):
            self.param["rprop"] = dict(
                lr=lr,
                etas=etas,
                step_sizes=step_sizes,
            )
            return self

        def SGD(self, lr=0.1, momentum=0, dampening=0, weight_decay=0, nesterov=False):
            self.param["sgd"] = dict(
                lr=lr,
                momentum=momentum,
                dampening=dampening,
                weight_decay=weight_decay,
                nesterov=nesterov,
            )
            return self

    @staticmethod
    def build():
        return OptimParam.OptimBuilder(OptimParam())

    def __repr__(self) -> str:
        return pp.pformat(self.optim_param)



"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""




"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""
import bisect
import os
import pprint as pp
import warnings
from functools import wraps
from typing import Any

import torch
from torch.optim.optimizer import Optimizer
from torch.utils.data.dataloader import DataLoader
from torch.utils.tensorboard import SummaryWriter

from .databundler import DataBundler
from .experiment import Experiment
from .logger import Logger
from .meter import AvgMeter
from .params import Params
from .plotter import Plotter
from .rndmanager import RndManager
from .saver import Saver


class BaseTrainer:
    _ignore_call_back = {"model_dict", "optim_dict",
                         "model_state_dict", "optim_state_dict",
                         "create_checkpoint_dict", "create_checkpoint_dict",
                         "iter_train_dataloader", "iter_eval_dataloader", "iter_test_dataloader",
                         "predict", "preprocess",
                         "add_callback"}

    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)

        def wrapper(func, _call_set: list):
            @wraps(func)
            def _newfunc(*args, **kwargs):
                for callback in _call_set:
                    if callback.enable:
                        callback.on_begin(self, func, self.params, *args, **kwargs)
                try:
                    _meter = func(*args, **kwargs)
                except BaseException as e:
                    _handles = [callback.on_exception(self, func, self.params, e, *args, **kwargs)
                                for callback in _call_set]
                    if any(_handles):
                        return None
                    else:
                        raise e

                for callback in _call_set:
                    if callback.enable:
                        callback.on_end(self, func, self.params, _meter, *args, **kwargs)
                return _meter

            return _newfunc

        self._callback_set = []
        self._callback_name_set = set()

        vars = dir(self)
        for name in vars:
            value = getattr(self, name)
            if name.startswith("_"):
                continue
            if callable(value):
                setattr(self, name, wrapper(value, self._callback_set))
        return self

    def __init__(self, params: Params):
        self.model_dict = {}
        self.optim_dict = {}
        self.databundler_dict = {}
        self.params = params
        self.train_epoch_toggle = False
        self.train_toggle = False
        self.device = torch.device(params.device)

    def initial_callback(self):
        pass

    def initial_trainer(self, params: Params):
        pass

    def initial_exp(self, exps_dir):
        exp_dir = os.path.join(exps_dir, self.params.get_exp_name())
        self.experiment = Experiment(exp_dir)
        self.experiment.start_exp()
        self.logger = Logger()
        self.logger.add_log_dir(self.experiment.hold_exp_part("logs", exts=[".log"]))
        self.saver = Saver(self.experiment.hold_exp_part("modules", exts=[".pth", ".json"]))
        self.plotter = Plotter(self.experiment.hold_exp_part("plot", exts=[".eps", ".jpeg", ".jpg",
                                                                           ".pdf", ".png", ".svg", ".tif", ".tiff"]))
        self.rnd = RndManager(self.experiment.hold_exp_part("rnd", exts=[".rnd"]))
        self.experiment.add_event_listener(self.plotter.dynamic_board, exts=[".bd"])
        self.writter = SummaryWriter(self.experiment.hold_exp_part("board", exts=[".bd"]), filename_suffix=".bd")
        self.initial_callback()
        self.initial_trainer(self.params)

    def train(self):
        params = self.params
        for eidx in range(params.eidx, params.epoch + 1):
            self.train_epoch(params.eidx, params)
            params.eidx = eidx + 1
            if self.train_toggle:
                self.train_toggle = False
                break

    def train_epoch(self, eidx, params):
        avgMeter = AvgMeter()
        for idx, batch_data in enumerate(self.iter_train_dataloader()):
            self.change_mode(True)
            meter = self.train_batch(eidx, idx, self.params.global_step, batch_data, params, self.device)
            avgMeter.update(meter)
            self.change_mode(False)

            params.global_step += 1
            params.idx = idx
            if self.train_epoch_toggle:
                self.train_epoch_toggle = False
                break
        return avgMeter

    def train_step(self, steps):
        param = self.params
        i = 0
        while steps > 0:
            for idx, data in enumerate(self.iter_train_dataloader()):
                meter = self.train_batch(0, idx, i, data, param, self.device)
                steps -= 1
                if steps <= 0:
                    return meter

    def train_on_batch(self, batch_data):
        self.train_batch(0, 0, 0, batch_data, self.params, self.device)

    def test(self):
        return self.test_eval_logic(self.iter_test_dataloader(), self.params)

    def eval(self):
        return self.test_eval_logic(self.iter_eval_dataloader(), self.params)

    def _regist_databundler(self, key, val):
        assert isinstance(val, (DataBundler, DataLoader))
        if isinstance(val, DataLoader):
            val = DataBundler().add(val)
        self.databundler_dict[key] = val

    def regist_databundler(self, train, eval, test):
        self.regist_train_databundler(train)
        self.regist_eval_databundler(eval)
        self.regist_test_databundler(test)

    def regist_eval_databundler(self, eval):
        self._regist_databundler("eval", eval)

    def regist_test_databundler(self, test):
        self._regist_databundler("tests", test)

    def regist_train_databundler(self, train):
        self._regist_databundler("train", train)

    def iter_train_dataloader(self) -> DataBundler:
        return self.databundler_dict["train"]

    def iter_eval_dataloader(self) -> DataBundler:
        return self.databundler_dict["eval"]

    def iter_test_dataloader(self) -> DataBundler:
        return self.databundler_dict["tests"]

    def _load_checkpoint_dict(self, res, strict, ignore_optim):
        if res is None:
            return False
        self.params.eidx = res["_eidx"]
        self.params.idx = res["_idx"]
        self.params.global_step = res["_global_step"]
        self.load_state_dict(res, strict)
        if not ignore_optim:
            self.load_optim_dict(res, strict)
        return True

    def load_keypoint(self, epoch, strict=True, ignore_optim=False):
        res = self.saver.load_keypoint(epoch)
        if self._load_checkpoint_dict(res, strict, ignore_optim):
            self.logger.info("checkpoint of {} epoch not found. Choice from:".format(epoch))
            self.logger.info(self.saver.find_keypoints())
        else:
            self.logger.info("load keypoint of {} epoch. ")
            extra = self.saver.load_checkpoint_info(epoch)
            if extra is not None:
                self.logger.info("Extra info:")
                self.logger.info(pp.pformat(extra))

    def load_checkpoint(self, epoch, strict=True, byindex=False, ignore_optim=False, not_exist_ok=False):
        res = self.saver.load_checkpoint(epoch, byindex)
        if res is None and not not_exist_ok:
            self.logger.info("file of epoch {} not found, only have {}".format(epoch, self.saver.find_checkpoints()))
            raise FileNotFoundError()

        if self._load_checkpoint_dict(res, strict, ignore_optim):
            self.logger.info("checkpoint of {} epoch not found. Choice from:".format(epoch))
            self.logger.info(self.saver.find_checkpoints())
        else:
            self.logger.info("load checkpoint of {} epoch. ")
            extra = self.saver.load_checkpoint_info(epoch, byindex)
            if extra is not None:
                self.logger.info("Extra info:")
                self.logger.info(pp.pformat(extra))

    def load_model(self, val, strict=True):
        self.load_state_dict(self.saver.load_model(val), strict)

    def save_keypoint(self, extra_info=None, replacement=False):
        state_dict = self.checkpoint_dict()
        self.saver.save_keypoint(state_dict["_eidx"], state_dict, extra_info, replacement)

    def save_checkpoint(self, extra_info=None, replacement=False):
        state_dict = self.checkpoint_dict()
        self.saver.save_checkpoint(state_dict["_eidx"], state_dict, extra_info, replacement)

    def save_model(self, extra_info=None):
        state_dict = self.state_dict()
        self.saver.save_model(self.params.eidx, state_dict, extra_info)

    def add_callback(self, callback):
        """
        添加一个回调函数
        :type callable,str
        :param callback:
        :return:
        """
        msg = None
        cb_name = callback.__class__.__name__
        if callback not in self._callback_set and cb_name in self._callback_name_set:
            msg = "Callback duplicate."
            callback.on_hook_failed(self, msg)

        if msg is not None:
            return False

        bisect.insort(self._callback_set, callback)
        self._callback_name_set.add(cb_name)
        callback.on_hooked(self, self.params)
        return True

    def remove_callback(self, callback):
        msg = None
        if callback not in self._callback_set:
            return False

        cb_name = callback.__class__.__name__
        self._callback_set.remove(callback)
        self._callback_name_set.remove(cb_name)
        return True

    '''module和optim的一部分方法集成'''

    def load_state_dict(self, state_dict, strict=True):
        for k in self.model_dict:
            if k in state_dict:
                self.model_dict[k].load_state_dict(state_dict[k], strict=strict)
            else:
                if strict:
                    raise KeyError(k)
                else:
                    warnings.warn("{} not found in state_dict".format(k))

    def load_optim_dict(self, state_dict, strict=True):
        for k in self.optim_dict:
            if k in state_dict:
                self.optim_dict[k].load_state_dict(state_dict[k])
            else:
                if strict:
                    raise KeyError(k)
                else:
                    warnings.warn("{} not found in state_dict".format(k))

    def load_one_state_dict(self, name, state_dict, strict=True):
        self.model_dict[name].load_state_dict(state_dict, strict)

    def state_dict(self):
        return {k: v.state_dict() for k, v in self.model_dict.items()}

    def one_state_dict(self, name):
        return self.model_dict[name].state_dict()

    def checkpoint_dict(self):
        res = self.state_dict()
        res.update(self.optim_state_dict())
        res["_eidx"] = self.params.eidx + 1
        res["_idx"] = self.params.idx
        res["_global_step"] = self.params.global_step
        return res

    def optim_state_dict(self):
        return {k: v.state_dict() for k, v in self.optim_dict.items()}

    def change_mode(self, train=True):
        for k, v in self.model_dict.items():
            if train:
                v.train()
            else:
                v.eval()

    def to(self, device):
        for k, v in self.model_dict.items():
            self.__setattr__(k, v.to(device))

    '''magic functions'''

    def __setattr__(self, name: str, value: Any) -> None:
        super().__setattr__(name, value)
        if isinstance(value, torch.device):
            pass
        elif isinstance(value, torch.nn.Module):
            self.model_dict[name] = value
        elif isinstance(value, Optimizer):
            self.optim_dict[name] = value

    # need to reimplement

    def train_batch(self, eidx, idx, global_step, batch_data, params: Params, device: torch.device):
        raise NotImplementedError()

    def test_eval_logic(self, dataloader, param: Params):
        raise NotImplementedError()

    def predict(self, xs):
        raise NotImplementedError()

    def estimate_memory(self):
        for _, v in self.model_dict.items():
            pass


class Trainer(BaseTrainer):
    def initial_callback(self):
        from .callbacks import EvalCallback, LoggerCallback
        ec = EvalCallback(1, 5)
        ec.hook(self)

        lc = LoggerCallback()
        lc.hook(self)

    def initial_trainer(self, params: Params):
        pass

    def predict(self, xs):
        pass

    def train_batch(self, eidx, idx, global_step, batch_data, params: Params, device: torch.device):
        pass

    def test_eval_logic(self, dataloader, param: Params):
        pass







"""
    Copyright (C) 2020 Shandong University

    This program is licensed under the GNU General Public License 3.0 
    (https://www.gnu.org/licenses/gpl-3.0.html). 
    Any derivative work obtained under this license must be licensed 
    under the GNU General Public License as published by the Free 
    Software Foundation, either Version 3 of the License, or (at your option) 
    any later version, if this derivative work is distributed to a third party.

    The copyright for the program is owned by Shandong University. 
    For commercial projects that require the ability to distribute 
    the code of this program as part of a program that cannot be 
    distributed under the GNU General Public License, please contact 
            
            sailist@outlook.com
             
    to purchase a commercial license.
"""
import torch


def classify(preds, labels, cacu_rate=False, topk=None):
    """
    用于分类的准确率
    :param preds: [batch,logits]
    :param labels: [labels,]
    :param cacu_rate: 计算正确率而不是计数
    :param topk: list(int) ，表明计算哪些topk
    :return:
        if cacu_rate:
            [topk_rate,...]
        else:
            total, [topk_count,...]
    """
    if topk is None:
        topk = (1,5)
    k = topk
    _, maxk = torch.topk(preds, max(*k), dim=-1)
    total = labels.size(0)
    test_labels = labels.view(-1, 1)  # reshape labels from [n] to [n,1] to compare [n,k]

    if cacu_rate:
        return [(test_labels == maxk[:, 0:i]).sum().item() / total for i in k]
    else:
        return total, [(test_labels == maxk[:, 0:i]).sum().item() for i in k]